/*----------------------------------------------------------------------
  File    : MWT.java
  Contents: find minimum weight triangulation of a pointgon
  Author  : Christian Borgelt
  History : 2005.02.18 file created
            2005.02.24 recursive functions completed
            2005.03.05 subproblem key format changed, functions adapted
            2005.03.09 changed to new processing scheme
            2005.03.23 precomputation of intersection table added
            2005.04.15 splitters are reused in branches of recursion
            2005.04.20 check of value returned from recursion added
            2005.05.11 prepared code for triangle-based algorithm
            2005.06.23 adapted to changes in Cutter
            2005.12.15 precomputation of more information added
            2005.12.16 finding valid edges made a separate function
            2005.12.28 adapted to abstract splitter class
            2006.01.07 precheck of holes on edges added
            2006.01.19 third subproblem splitter (Combined) added
            2006.01.21 check of return code of trie.add added
            2006.01.27 second path-based algorithm added
            2006.02.06 adapted to new Splitter interface method split
            2006.02.08 adapted to fourth subproblem splitter class
            2006.02.18 simple version of greedy triangulation added
            2006.02.20 bug in mode mask fixed (PRECHECK not included)
            2016.04.07 StringBuffer replaced by StringBuilder
----------------------------------------------------------------------*/
package pointgon;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.Arrays;

/*--------------------------------------------------------------------*/
/** Class for the minimum weight triangulation of a pointgon.
 *  @since  2005.02.18 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class MWT implements Runnable, Serializable {

  private static final long serialVersionUID = 0x00010000L;

  /** block size for array enlargement */
  private static final int BLKSIZE   = 32;
  /** splitter identifier: traingle */
  public  static final int TRIANGLE  =  0;
  /** splitter identifier: path */
  public  static final int PATH      =  1;
  /** splitter identifier: combined 1 */
  public  static final int COMB_1    =  2;
  /** splitter identifier: combined 2 */
  public  static final int COMB_2    =  3;
  /** splitter identifier: greedy */
  public  static final int GREEDY    =  4;
  /** processing mode: algorithm mask */
  public  static final int ALGOMASK  = 0x07;
  /** processing mode: precheck flag */
  public  static final int PRECHECK  = 0x10;
  /** processing mode: hole distribution flag */
  public  static final int DISTHOLES = 0x20;
  /** processing mode: reverse key flag */
  public  static final int REVKEY    = 0x40;

  /** perimeter vertices of pointgon */
  protected Vertex[]   verts;
  /** order of the perimeter vertices */
  protected int        dir;
  /** hole vertices (sorted by x-coord.) */
  protected Vertex[]   holes;
  /** table of edge weights */
  protected double[][] wgts;
  /** flag table for valid triangles */
  protected char[][][] trtab;
  /** processing mode (e.g. DISTHOLES) */
  protected int        mode;
  /** vector of subproblem splitters */
  private   Splitter[] splits;
  /** trie for subprob. solution lookup */
  private   Trie       trie;
  /** vector to collect solution edges */
  private   Edge[]     edges;
  /** current number of edges in vector */
  private   int        cnt;
  /** weight of best triangulation */
  private   double     best;
  /** whether thread has been stopped */
  private   boolean    stopped;
  /** maximal recursion depth */
  private   int        maxdp;
  /** number of checked triangles */
  private   long       tricnt;
  /** number of stored subproblems */
  private   long       subcnt;
  /** number of valid subproblem splits */
  private   long       splcnt;
  /** execution time of search */
  private   long       time;

  /*------------------------------------------------------------------*/
  /** Get the processing mode.
   *  @return the processing mode
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int getMode ()
  { return this.mode; }

  /*------------------------------------------------------------------*/
  /** Set the processing mode.
   *  @param  mode   the processing mode
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setMode (int mode)
  { this.mode = mode; }

  /*------------------------------------------------------------------*/
  /** Get the perimeter vertices of the pointgon.
   *  @return the perimeter vertices of the pointgon
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Vertex[] getVertices ()
  { return this.verts; }

  /*------------------------------------------------------------------*/
  /** Get the holes of the pointgon.
   *  @return the holes of the pointgon
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Vertex[] getHoles ()
  { return this.holes; }

  /*------------------------------------------------------------------*/
  /** Get the weight of the best triangulation.
   *  @return the weight of the best triangulation
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getWeight ()
  { return this.best; }

  /*------------------------------------------------------------------*/
  /** Get the edges of the best triangulation.
   *  @return the edges of the best triangulation
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Edge[] getEdges ()
  { return this.edges; }

  /*------------------------------------------------------------------*/
  /** Get the number of triangles of the best triangulation.
   *  @return the number of triangles of the best triangulation
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public long getTriangles ()
  { return this.tricnt; }

  /*------------------------------------------------------------------*/
  /** Get the number of subproblems visited.
   *  @return the number of subproblems visited
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public long getSubprobs ()
  { return this.subcnt; }

  /*------------------------------------------------------------------*/
  /** Get the number of splits considered.
   *  @return the number of splits considered
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public long getSplits ()
  { return this.splcnt; }

  /*------------------------------------------------------------------*/
  /** Get the maximum depth of the recursion.
   *  @return the maximum depth of the recursion
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public long getDepth ()
  { return this.maxdp; }

  /*------------------------------------------------------------------*/
  /** Get the execution time of the search.
   *  @return the execution time of the search
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public long getTime ()
  { return this.time; }

  /*------------------------------------------------------------------*/
  /** Check whether an edge is inside the pointgon.
   *  @param  src  the source      vertex of the edge
   *  @param  dst  the destination vertex of the edge
   *  @return whether the edge is inside the pointgon
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean isInside (Vertex src, Vertex dst)
  {                             /* --- check whether edge is inside */
    int     i, n;               /* loop variable, number of vertices */
    Vertex  p, q;               /* vertices adjacent to source */
    boolean rgt, lft;           /* flags for point positions */

    /* --- check whether edge is inside --- */
    n = this.verts.length;      /* get number of perimeter vertices */
    if ((src.id >= n)           /* if there is a perimeter vertex, */
    &&  (dst.id <  n)) {        /* source should be on the perimeter */
      p = dst; dst = src; src = p; }
    if (src.id < n) {           /* if source is on the perimeter */
      q = this.verts[(src.id +n +this.dir) % n];
      p = this.verts[(src.id +n -this.dir) % n];
      if ((dst != p)            /* if the edge to test */
      &&  (dst != q)) {         /* is not a perimeter edge */
        lft = dst.isLeftOf(src, q);  /* position relative */
        rgt = dst.isRightOf(src,p);  /* to incident edges */
        if (src.isLeftOf(q,p)){ /* pointgon is convex  at source */
          if (!lft || !rgt) return false; }
        else {                  /* pointgon is concave at source */
          if (!lft && !rgt) return false; }
      }                         /* edge must lead "to the inside" */
    }                           /* (must be between incident edges) */

    /* --- check intersection with perimeter --- */
    p = this.verts[0];          /* get first vertex for wrap around */
    for (i = this.verts.length; --i >= 0; ) {
      q = p; p = this.verts[i]; /* traverse the perimeter edges */
      if ((p != src) && (p != dst)   /* skip incident */
      &&  (q != src) && (q != dst)   /* perimeter edges */
      &&  p.isectsX(q, src, dst))    /* if the perimeter edge */
        return false;           /* intersects the given edge, */
    }                           /* abort the function */
    return true;                /* return 'edge is fully inside' */
  }  /* isInside() */

  /*------------------------------------------------------------------*/
  /** Check whether an edge has a hole.
   *  @param  src  the source      vertex of the edge
   *  @param  dst  the destination vertex of the edge
   *  @return whether the edge has a hole
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean hasHole (Vertex src, Vertex dst)
  {                             /* --- check whether edge has a hole */
    int    i;                   /* loop variable */
    Vertex h;                   /* to traverse the holes */

    for (i = this.holes.length; --i >= 0; ) {
      h = this.holes[i];        /* traverse the hole vertices */
      if ((h != src) && (h != dst) && h.isPartOf(src, dst))
        return true;            /* if a hole is part of the edge, */
    }                           /* abort with 'edge contains a hole' */
    return false;               /* return 'edge contains no hole' */
  }  /* hasHole() */

  /*------------------------------------------------------------------*/
  /** Precompute valid edges.
   *  @return the created weight table
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private double[][] validEdges ()
  {                             /* --- precompute valid edges */
    int        i, k, n;         /* loop variables, number of vertices */
    Vertex     p, q;            /* to traverse the vertices */
    double[][] tab;             /* created weight table */

    n   = this.verts.length;    /* get number of perimeter vertices */
    i   = n +this.holes.length; /* and the number of holes */
    tab = new double[i][i];     /* create a table of edge weights */
    while (--i > 0) {           /* traverse all vertices */
      p = (i < n) ? this.verts[i] : this.holes[i-n];
      for (k = i; --k >= 0; ) { /* traverse all inner edges */
        q = (k < n) ? this.verts[k] : this.holes[k-n];
        if (this.isInside(q, p) /* if the edge is inside the pointgon */
        && !this.hasHole(q, p)) /* and does not contain a hole */
          tab[p.id][q.id] = tab[q.id][p.id] = p.distance(q);
        else                    /* if the edge is invalid */
          tab[p.id][q.id] = tab[q.id][p.id] = -1.0;
      }                         /* clear/invalidate the edge weight */
    }                           /* (use both index combinations) */
    return tab;                 /* return the created weight table */
  }  /* validEdges() */

  /*------------------------------------------------------------------*/
  /** Precompute valid triangles.
   *  @return the created flag table
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private char[][][] validTriangles ()
  {                             /* --- precompute valid triangles */
    int        i, j, k, l, n;   /* loop variables, number of vertices */
    Vertex     a, b, c, h;      /* to traverse the vertices */
    char[][][] tab;             /* created flag table */

    n   = this.verts.length;    /* get number of perimeter vertices */
    i   = n +this.holes.length; /* and the number of holes */
    tab = new char[i][i][i];    /* create the flag table */
    while (--i > 0) {           /* traverse possible first corners */
      a = (i < n) ? this.verts[i] : this.holes[i-n];
      for (j = i; --j > 0; ) {  /* traverse possible second corners */
        b = (j < n) ? this.verts[j] : this.holes[j-n];
        if (this.wgts[a.id][b.id] < 0)
          continue;             /* check first triangle side */
        for (k = j; --k >= 0;){ /* traverse possible third corners */
          c = (k < n) ? this.verts[k] : this.holes[k-n];
          if ((this.wgts[a.id][c.id] < 0)
          ||  (this.wgts[b.id][c.id] < 0))
            continue;           /* check two other triangle sides */
          for (l = this.holes.length; --l >= 0; ) {
            h = this.holes[l];  /* traverse the holes */
            if ((h != a) && (h != b) && (h != c)
            &&  h.isInsideX(a, b, c))
              break;            /* check whether a hole */
          }                     /* is inside the triangle */
          if (l >= 0) continue; /* if triangle is not empty, skip it */
          tab[a.id][b.id][c.id] = tab[a.id][c.id][b.id] =
          tab[b.id][a.id][c.id] = tab[b.id][c.id][a.id] =
          tab[c.id][a.id][b.id] = tab[c.id][b.id][a.id] = 127;
        }                       /* set flag for a valid triangle */
      }                         /* for all six combinations */
    }                           /* of the vertex indices */
    return tab;                 /* return the created flag table */
  }  /* validTriangles() */

  /*------------------------------------------------------------------*/
  /** Find a greedy triangulation.
   *  @return whether the search was successfully completed
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean greedy ()
  {                             /* --- find greedy triangulation */
    int    i, k, n, m;          /* loop variables, edge counter */
    Vertex p, q;                /* to traverse the vertices */
    Edge   e;                   /* to traverse the edges */
    Edge[] eds;                 /* vector of valid/selected edges */
    double w;                   /* buffer for (sum of) edge weight(s) */

    this.best = -1.0;           /* initialize the result (weight) */
    n = this.verts.length;      /* get number of perimeter vertices */
    m = 0;                      /* traverse all vertex pairs */
    for (i = n +this.holes.length; --i > 0; )
      for (k = i; --k >= 0; )   /* count the valid edges */
        if (this.wgts[i][k] >= 0) m++;
    eds = new Edge[m-n];        /* create a vector for the edges */
    m = 0;                      /* traverse all vertex pairs again */
    for (i = n +this.holes.length; --i > 0; ) {
      p = (i < n) ? this.verts[i] : this.holes[i-n];
      for (k = (i < n) ? i-1 : i; --k > 0; ) {
        q = (k < n) ? this.verts[k] : this.holes[k-n];
        w = this.wgts[p.id][q.id];
        if (w >= 0) eds[m++] = new Edge(p, q, w);
      }                         /* create the valid edges and */
      w = this.wgts[p.id][0];   /* store them in the vector */
      if ((w >= 0) && (i > 1) && (i != n-1))
        eds[m++] = new Edge(p, this.verts[0], w);
    }                           /* (exclude the perimeter edges) */
    Arrays.sort(eds);           /* sort the valid edges */
    for (w = k = 0, i = n; --i >= 0; k = i)
      w += this.wgts[i][k];     /* sum weight of perimeter edges */
    for (i = n = 0; i < m; i++) {
      e = eds[i];               /* traverse the valid edges */
      if (this.stopped)         /* check for an external abort */
        return false;           /* (if running as a separate thread) */
      for (k = n; --k >= 0; ) { /* traverse the selected edges */
        if (!e.meets(eds[k]) && e.isects(eds[k]))
          break;                /* check for a proper intersection */
      }
      if (k >= 0) continue;     /* if there is one, skip this edge */
      eds[n++] = e;             /* otherwise add the edge */
      w += e.wgt;               /* to the selected edges */
    }                           /* and sum the egde weight */
    this.edges = new Edge[n];   /* create vector of selected edges */
    System.arraycopy(eds, 0, this.edges, 0, n);
    this.best = w;              /* set the total edge weight */
    return true;                /* get the time and return 'ok' */
  }  /* greedy() */

  /*------------------------------------------------------------------*/
  /** Report an abnormal situation.
   *  @param  key      the subproblem key
   *  @param  cnt      the subproblem size
   *  @param  holes    the holes of the subproblem
   *  @param  holecnt  the number of holes of the subproblem
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void abnormal (Vertex[] key,   int cnt,
                         Vertex[] holes, int holecnt)
  {                             /* --- report an abnormal situation */
    System.err.println("abnormal situation detected!");
    System.err.println("vertices (" +this.verts.length +" perimeter):");
    for (int i = 0; i < cnt; i++) {
      if (key[i] != null) System.err.print(" " +key[i].id);
      System.err.println(" " +key[i]);
    }                           /* print subproblem key */
    System.err.println("holes    (" +this.holes.length +" initially):");
    for (int i = 0; i < holecnt; i++)
      System.err.println(" " +holes[i].id +" " +holes[i]);
  }  /* abnormal() */           /* print list of holes */

  /* The above function reports a subproblem for which an abnormal */
  /* situation was detected. This is only for debugging purposes,  */
  /* as such an abnormal situation indicates a bug in the program. */

  /*------------------------------------------------------------------*/
  /** Create a processing mode dependent splitter.
   *  @return the processing mode dependent splitter
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private Splitter splitter ()
  {                             /* --- create mode-dependent splitter */
    switch (this.mode & ALGOMASK) {
      case PATH  : return new Path     (this);
      case COMB_1: return new Combined1(this);
      case COMB_2: return new Combined2(this);
      default    : return new Triangle (this);
    }                           /* return proper subproblem splitter */
  }  /* splitter() */

  /*------------------------------------------------------------------*/
  /** Recursive solution procedure.
   *  @param  key      the subproblem key
   *  @param  cnt      the subproblem size
   *  @param  holes    the holes of the subproblem
   *  @param  holecnt  the number of holes of the subproblem
   *  @param  depth    the recursion depth
   *  @return the weight of the best solution of the subproblem
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private double recurse (Vertex[] key,   int cnt,
                          Vertex[] holes, int holecnt, int depth)
  {                             /* --- recursive solution procedure */
    double   lft, rgt;          /* weights of the subproblems */
    double   wgt, min;          /* current and best weight */
    Splitter split;             /* subproblem splitter */
    Vertex[] tri;               /* cut out triangle (if any) */

    if (this.stopped)           /* check for an external abort */
      return -1;                /* (if running as a separate thread) */
    if (depth > this.maxdp)     /* if new maximal depth reached, */
      this.maxdp = depth;       /* update the recursion depth */
    this.subcnt++;              /* count the subproblem */

    /* --- check for a triangle without holes --- */
    split = this.splits[depth]; /* get the level subproblem splitter */
    if (split == null) this.splits[depth] = split = this.splitter();
    split.init(key, cnt, holes, holecnt);
    // split.show(depth);          /* show subproblem (debugging) */
    wgt = split.empty();        /* initialize the subproblem splitter */
    if (wgt >= 0.0) {           /* if subproblem is an empty triangle */
      this.trie.add(key, cnt, wgt, null);
      this.tricnt++;            /* store the solution in the trie, */
      return wgt;               /* count the empty triangle, and */
    }                           /* return the weight of the triangle */

    /* --- split into subproblems and process them --- */
    depth++;                    /* increment the recursion depth */
    min = Double.MAX_VALUE;     /* init. the weight of the best split */
    while (split.next()) {      /* traverse splits into subproblems */
      this.splcnt++;            /* count the created subproblem split */
      split.split();            /* and split the current pointgon */
      tri = split.triangle();   /* get a possible cut triangle */
      if (tri != null) {        /* if there is such a triangle */
        wgt = this.trie.getWeight(tri, 3);
        if (wgt < 0) {          /* if the triangle is not known yet */
          wgt = this.wgts[tri[0].id][tri[1].id]
              + this.wgts[tri[0].id][tri[2].id]
              + this.wgts[tri[1].id][tri[2].id];
          this.trie.add(tri, 3, wgt, null);
          this.tricnt++;        /* store the triangle and its weight */
        }                       /* and count the empty triangle */
      }
      wgt = split.weight();     /* look up subproblems in the trie */
      rgt = (split.rgtcnt <= 0) /* get solution of right subproblem */
          ? 0 : this.trie.getWeight(split.rgtkey, split.rgtcnt);
      lft = (split.lftcnt <= 0) /* get solution of left  subproblem */
          ? 0 : this.trie.getWeight(split.lftkey, split.lftcnt);
      if ((rgt < 0)             /* if the solution of one or both */
      ||  (lft < 0)) {          /* of the subproblems is unknown, */
        // if (((lft > 0) ? lft : 0) +((rgt > 0) ? rgt : 0) +wgt > min)
        //   continue;
        split.distribute();     /* distribute the hole vertices */
        if (rgt < 0) {          /* process the right subproblem */
          rgt = this.recurse(split.rgtkey,   split.rgtcnt,
                             split.rgtholes, split.rgtholecnt, depth);
          if (rgt < 0) return rgt;
        }                       /* if the recursion failed, abort */
        if (lft < 0) {          /* process the left subproblem */
          // if (rgt +wgt > min) continue;
          lft = this.recurse(split.lftkey,   split.lftcnt,
                             split.lftholes, split.lftholecnt, depth);
          if (lft < 0) return lft;
        }                       /* if the recursion failed, abort */
      }                         /* solve the (unsolved) subproblem(s) */
      wgt = lft +rgt +wgt;      /* combine the subproblem solutions */
      if (wgt < min) {          /* if the current solution is better */
        min = wgt; split.store(); }
    }                           /* copy the path and note its weight */
    /* Note that in the above loop the holes are distributed only   */
    /* if the solution to a subproblem is not known. Otherwise the  */
    /* distribution is not necessary and therefore not carried out. */

    /* --- store the subproblem solution --- */
    if ((min >= Double.MAX_VALUE)  /* check for abnormal situation */
    ||  (this.trie.add(key, cnt, min, split.retrieve()) < 0)) {
      this.abnormal(key, cnt, holes, holecnt);
      this.stopped = true; return -1;
    }                           /* stop the computation and abort */
    /* It should not happen that the subproblem is unsolved, because */
    /* this would indicate an implementation bug. Nevertheless, this */
    /* is checked here to get a proper error message in this case.   */
    return min;                 /* store the solution in the trie */
  }  /* recurse() */            /* and return its weight */

  /*------------------------------------------------------------------*/
  /** Collect the edges of the minimum weight triangulation.
   *  @param  key      the subproblem key
   *  @param  cnt      the subproblem size
   *  @param  depth    the recursion depth
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void collect (Vertex[] key, int cnt, int depth)
  {                             /* --- collect edges of MWT */
    int      n, vsz;            /* number of edges, vector size */
    Object   spobj;             /* splitting path/triangle */
    Splitter split;             /* subproblem splitter */
    Edge[]   buf, e;            /* buffers for edge vectors */

    spobj = this.trie.getObject(key, cnt);
    if (spobj == null)          /* if there is no stored split, */
      return;                   /* there are no inner edges */
    split = this.splits[depth]; /* get the level subproblem splitter */
    if (split == null) this.splits[depth] = split = this.splitter();
    split.init(key,cnt,spobj);  /* init. the subproblem splitter */
    // split.show(depth);          /* show subproblem (debugging) */
    split.split();              /* and split the current pointgon */
    e   = split.edges();        /* get the edges of the best split */
    n   = this.cnt +e.length;   /* compute the new number of edges */
    vsz = this.edges.length;    /* get the edge vector size */
    if (n >= vsz) {             /* if the edge vector is too small */
      vsz += (vsz > BLKSIZE) ? vsz >> 1 : BLKSIZE;
      if (vsz < n) vsz = n;     /* compute new vector size and */
      buf = new Edge[vsz];      /* allocate an new edge vector */
      System.arraycopy(this.edges, 0, buf, 0, this.cnt);
      this.edges = buf;         /* copy the existing edges */
    }                           /* and set the new vector */
    System.arraycopy(e, 0, this.edges, this.cnt, e.length);
    this.cnt += e.length;       /* copy the edges to the solution */
    depth++;                    /* increment the recursion depth */
    if (split.lftcnt > 0)       /* rec. process left  subproblem */
      this.collect(split.lftkey, split.lftcnt, depth);
    if (split.rgtcnt > 0)       /* rec. process right subproblem */
      this.collect(split.rgtkey, split.rgtcnt, depth);
  }  /* collect() */            /* collect edges recursively */

  /*------------------------------------------------------------------*/
  /** Solve minimum weight triangulation problem.
   *  @return whether a solution was found
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean solve ()
  {                             /* --- solve minimum weight triang. */
    int      i, k, n;           /* loop variables, number of vertices */
    Vertex   v, p, q;           /* buffers for some vertices */
    Vertex[] key;               /* key of initial pointgon */
    Edge[]   buf;               /* buffer for vector reallocation */

    this.time    = System.currentTimeMillis();
    this.splcnt  = this.subcnt = this.tricnt = 0;
    this.stopped = false;       /* get time and init. statistics */

    /* --- determine vertex order and leftmost vertex --- */
    n = this.verts.length;      /* get number of perimeter vertices */
    v = this.verts[i = n-1];    /* start at the last vertex and */
    while (--i >= 0) {          /* traverse the perimeter vertices */
      if (this.verts[i].compareTo(v) <= 0)
        v = this.verts[i];      /* find the leftmost perimeter vertex */
    }                           /* (anchor for initial pointgon) */
    p = this.verts[(v.id    +1) % n];
    q = this.verts[(v.id +n -1) % n];
    this.dir = ((p.x -v.x) *(q.y -v.y) < (p.y -v.y) *(q.x -v.x))
             ? -1 : +1;         /* determine the vertex order */

    /* --- greedy triangulation --- */
    this.wgts = this.validEdges();   /* precompute edge weights */
    if ((this.mode & ALGOMASK) == GREEDY) {
      this.greedy();            /* compute greedy triangulation */
      this.time = System.currentTimeMillis() -this.time;
      return (this.best >= 0);  /* get execution time and return */
    }                           /* whether a solution was found */

    /* --- create key of initial pointgon --- */
    key    = new Vertex[3];     /* create initial key vector */
    key[0] = v;                 /* set the subproblem anchor */
    key[1] = (this.dir > 0) ? p : q;  /* and the ccw. vertex */
    key[2] = null;              /* and the perimeter section */

    /* --- solve minimum weight triangulation problem --- */
    this.trtab  = ((this.mode & PRECHECK) == 0)  /* precompute the */
                ? null : this.validTriangles();  /* valid triangles */
    i = this.holes.length;      /* create subproblem splitter vector */
    this.splits = new Splitter[n+i+i];   /* and trie for subproblems */
    this.trie   = new Trie(n, i, (this.mode & REVKEY) != 0);
    this.maxdp  = 0;            /* start the recursive search */
    this.best   = this.recurse(key, key.length, this.holes, i, 0);

    /* --- collect (inner) edges of solution --- */
    if (this.best >= 0) {       /* if the search succeeded */
      this.edges = new Edge[BLKSIZE];
      this.cnt   = 0;           /* collect edges of triangulation */
      this.collect(key, key.length, 0);
      if (this.cnt < this.edges.length) {
        buf = new Edge[this.cnt];
        System.arraycopy(this.edges, 0, buf, 0, this.cnt);
        this.edges = buf;       /* if the edge vector is not full, */
      }                         /* shrink the edge vector */
    }                           /* to the needed size */

    /* --- trie statistics for profiling --- */
    // int[] stats = this.trie.getStats();
    // System.err.println("trie statistics:");
    // System.err.println("number of nodes      : " +stats[0]);
    // System.err.println("number of used nodes : " +stats[1]);
    // System.err.println("number of inner nodes: " +stats[2]);
    // System.err.println("number of leaves     : " +stats[3]);
    // System.err.println("number of pointers   : " +stats[4]);

    /* --- clean up --- */
    this.time   = System.currentTimeMillis() -this.time;
    this.trie   = null;         /* get the execution time and */
    this.splits = null;         /* and "delete" the data structures */
    return (this.best >= 0);    /* return whether solution was found */
  }  /* solve() */

  /*------------------------------------------------------------------*/
  /** Run minimum weight triangulation problem solving.
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void run ()
  { this.solve(); }
    
  /*------------------------------------------------------------------*/
  /** Stop the minimum weight triangulation problem solving (abort).
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void stop ()
  { this.stopped = true; }

  /*------------------------------------------------------------------*/
  /** Create a string description of the minimum weight triangulation.
   *  @return the created string description
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String toString ()
  {                             /* --- create a string description */
    StringBuilder s = new StringBuilder("MWT = {\n");
    for (int i = 0; i < this.edges.length; i++)
      s.append("  " +this.edges[i] +",\n");
    s.append("};\n");           /* terminate the description */
    return s.toString();        /* return the created string */
  }  /* toString() */

  /*------------------------------------------------------------------*/
  /** Create a minimum weight triangulation solver.
   *  @param  pgon  the pointgon to triangulate
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public MWT (Pointgon pgon)
  {                             /* --- minimum weight triangulation */
    this.verts = pgon.verts;    /* note the perimeter vertices */
    if (pgon.holes == null)     /* if there are no holes, */
      this.holes = new Vertex[0];  /* set an empty vector */
    else {                      /* if there are holes */
      this.holes = new Vertex[pgon.holes.length];
      System.arraycopy(pgon.holes, 0, this.holes, 0, pgon.holes.length);
      Arrays.sort(this.holes, 0, this.holes.length);
    }                           /* sort a copy of the holes */
    this.best  = -1;            /* to preserve original pointgon */
    this.edges = null;          /* init. weight and edges to indicate */
    this.mode  = TRIANGLE|PRECHECK;  /* that the MWT is not known yet */
  }  /* MWT() */

  /*------------------------------------------------------------------*/
  /** Main program for command line invocation.
   *  @param  args  the command line arguments
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function for testing */
    int      i, mode;           /* loop variable, processing mode */
    Pointgon pgon;              /* pointgon (polygon with holes) */
    MWT      mwt;               /* minimum weight triangulation */
    Edge[]   edges;             /* triangulation edges */
    String   algo;              /* name of the algorithm used */

    if (args.length <= 0) {     /* if no file name is given */
      System.err.println("usage: MWT <file> [options]"); return; }
    try {                       /* parse the file */
      pgon = new Pointgon(args[0].equals("-") ? System.in
                        : new FileInputStream(args[0])); }
    catch (IOException ioe) {   /* catch and print exception */
      System.err.println(ioe.getMessage()); return; }
    if (!pgon.isValid()) {      /* check whether pointgon is valid */
      System.err.println("error: pointgon is invalid."); return; }
    mwt  = new MWT(pgon);       /* create an MWT object and get */
    mode = COMB_2|PRECHECK|REVKEY;   /* default processing mode */
    for (i = args.length; --i > 0; ) {
      if      (args[i].equalsIgnoreCase("triangle"))
        mode = TRIANGLE | (mode & ~ALGOMASK);
      else if (args[i].equalsIgnoreCase("path"))
        mode = PATH     | (mode & ~ALGOMASK);
      else if (args[i].equalsIgnoreCase("comb_1")
      ||       args[i].equalsIgnoreCase("comb1"))
        mode = COMB_1   | (mode & ~ALGOMASK);
      else if (args[i].equalsIgnoreCase("comb_2")
      ||       args[i].equalsIgnoreCase("comb2"))
        mode = COMB_2   | (mode & ~ALGOMASK);
      else if (args[i].equalsIgnoreCase("greedy"))
        mode = GREEDY   | (mode & ~ALGOMASK);
      else if (args[i].equals("+d")) mode |=  DISTHOLES;
      else if (args[i].equals("-d")) mode &= ~DISTHOLES;
      else if (args[i].equals("+p")) mode |=  PRECHECK;
      else if (args[i].equals("-p")) mode &= ~PRECHECK;
      else if (args[i].equals("+r")) mode |=  REVKEY;
      else if (args[i].equals("-r")) mode &= ~REVKEY;
    }                           /* evaluate the program options */
    mwt.setMode(mode);          /* set the processing mode and */
    mwt.solve();                /* find the MWT of the pointgon */
    edges = mwt.getEdges();     /* get the triangulation edges */
    pgon.setEdges(edges);       /* and set them in the pointgon */
    System.out.println(pgon);   /* print the pointgon with edges */
    switch (mode & ALGOMASK) {  /* evaluate the algorithm code */
      case TRIANGLE: algo = "triangle"; break;
      case PATH:     algo = "path";     break;
      case COMB_1:   algo = "comb_1";   break;
      case COMB_2:   algo = "comb_2";   break;
      default:       algo = "greedy";   break;
    }                           /* get the algorithm name */
    algo += (((mode & PRECHECK)  != 0) ? ", precheck"    : "")
         +  (((mode & DISTHOLES) != 0) ? ", dist. holes" : "")
         +  (((mode & REVKEY)    != 0) ? ", reverse key" : "");
    System.out.println("search statistics:"
      +"\nMWT algorithm used   : " +algo
      +"\nmax. recursion depth : " +mwt.getDepth()
      +"\nnumber of triangles  : " +mwt.getTriangles()
      +"\nnumber of subproblems: " +mwt.getSubprobs()
      +"\nnumber of splits     : " +mwt.getSplits()
      +"\ntotal search time    : " +(mwt.getTime()/1000.0) +"s"
      +"\n\nsolution statistics:"
      +"\nperimeter vertices   : " +mwt.getVertices().length
      +"\nhole vertices        : " +mwt.getHoles().length
      +"\nnumber of edges      : " +((edges != null) ? edges.length : 0)
      +"\ntriangulation weight : " +mwt.getWeight());
  }  /* main() */

}  /* class MWT */
