/*----------------------------------------------------------------------
  File    : PGPanel.java
  Contents: pointgon viewer panel
  Author  : Christian Borgelt
  History : 2005.02.20 file created from file DTPanel.java
            2005.03.19 functions findVertex and findEdge added
----------------------------------------------------------------------*/
package pointgon;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

/*--------------------------------------------------------------------*/
/** Class for a pointgon viewer panel.
 *  @since  2005.02.20 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class PGPanel extends JPanel {

  private static final long serialVersionUID = 0x00010000L;

  /** pointgon to visualize */
  private Pointgon pgon;
  /** x-offset for translation */
  private double   xoff;
  /** y-offset translation */
  private double   yoff;
  /** scaling factor */
  private double   scale;

  /*------------------------------------------------------------------*/
  /** Create a pointgon viewer panel.
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PGPanel ()
  {                             /* --- create a pointgon panel */
    this.pgon  = null;          /* there is no pointgon yet */
    this.xoff  = this.yoff = 0; /* initialize offset */
    this.scale = 16.0;          /* and scale */
  }  /* PGPanel() */

  /*------------------------------------------------------------------*/
  /** Create a pointgon viewer panel.
   *  @param  pgon  the pointgon to display
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PGPanel (Pointgon pgon)
  { this(); this.setPointgon(pgon); }

  /*------------------------------------------------------------------*/
  /** Get the pointgon of the viewer panel.
   *  @return the pointgon of the viewer panel
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Pointgon getPointgon ()
  { return this.pgon; }

  /*------------------------------------------------------------------*/
  /** Set the pointgon of the viewer panel.
   *  @param  pgon  the pointgon of the viewer panel
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setPointgon (Pointgon pgon)
  {                             /* --- set a new pointgon */
    this.pgon = pgon;           /* store the new pointgon */
    if (this.pgon == null)      /* if no pointgon, set default size */
      this.setPreferredSize(new Dimension(400, 400));
    else {                      /* if a new pointgon was set */
      this.xoff = this.pgon.getX();
      this.yoff = this.pgon.getY();
      this.setScale(16.0);      /* set default offset and scale */
    }
    this.revalidate();          /* adapt the enclosing scroll pane */
    this.repaint();             /* and redraw the pointgon */
  }  /* setPointgon() */

  /*------------------------------------------------------------------*/
  /** Get the scale of the pointgon viewer panel.
   *  @return the scale of the viewer panel
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getScale ()
  { return this.scale; }

  /*------------------------------------------------------------------*/
  /** Set the scale of the pointgon viewer panel.
   *  @param  scale  the scale of the viewer panel
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setScale (double scale)
  {                             /* --- rescale the display */
    Dimension d;                /* preferred size of panel */

    this.scale = scale;         /* set new scaling factor */
    d = new Dimension();        /* compute new preferred size */
    d.width  = (int)(this.pgon.getWidth()  *this.scale +16.5);
    d.height = (int)(this.pgon.getHeight() *this.scale +16.5);
    this.setPreferredSize(d);   /* set new preferred size */
  }  /* setScale() */

  /*------------------------------------------------------------------*/
  /** Draw the panel contents.
   *  @param  g   the graphics context
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void paint (Graphics g)
  {                             /* --- (re)paint the whole panel */
    int       i;                /* loop variable */
    Dimension d;                /* (preferred) size of panel */
    int       w, h;             /* size of background rectangle */
    int       x, y, ox, oy;     /* coordinates of points */
    Vertex    v;                /* to traverse the vertices */
    Edge      e;                /* to traverse the edges */

    d = this.getSize();         /* get the (preferred) panel size */
    w = d.width; h = d.height;  /* (whichever is larger) */
    d = this.getPreferredSize();
    if (d.width  > w) w = d.width;
    if (d.height > h) h = d.height;
    g.setColor(Color.white);    /* set the background color */
    g.fillRect(0, 0, w, h);     /* draw the background */
    if (this.pgon == null)      /* if there is no pointgon, */
      return;                   /* abort the function */
    w = 8; h = d.height -8;     /* coordinates of origin */

    /* --- draw the edges --- */
    if (this.pgon.edges != null) {
      g.setColor(Color.red);    /* edges are drawn in red */
      for (i = this.pgon.edges.length; --i >= 0; ) {
        e  = this.pgon.edges[i];/* traverse the edges */
        ox = (int)(w +this.scale *(e.src.x -this.xoff) +0.5);
        oy = (int)(h -this.scale *(e.src.y -this.yoff) +0.5);
        x  = (int)(w +this.scale *(e.dst.x -this.xoff) +0.5);
        y  = (int)(h -this.scale *(e.dst.y -this.yoff) +0.5);
        g.drawLine(ox, oy, x, y);
      }                         /* draw a line from the source to */
    }                           /* the destination of the edge */

    /* --- draw the perimeter --- */
    g.setColor(Color.black);    /* perimeter is drawn in black */
    v  = this.pgon.verts[0];    /* get the first perimeter vertex */
    ox = (int)(w +this.scale *(v.x -this.xoff) +0.5);
    oy = (int)(h -this.scale *(v.y -this.yoff) +0.5);
    for (i = this.pgon.verts.length; --i >= 0; ) {
      v = this.pgon.verts[i];   /* traverse the perimeter vertices */
      x = (int)(w +this.scale *(v.x -this.xoff) +0.5);
      y = (int)(h -this.scale *(v.y -this.yoff) +0.5);
      g.drawLine(ox, oy, x, y);
      g.fillRect(x-1, y-2, 3, 5);
      g.fillRect(x-2, y-1, 5, 3);
      ox = x; oy = y;           /* draw connecting line and circle, */
    }                           /* then go to the next vertex */

    /* --- draw the holes --- */
    for (i = this.pgon.holes.length; --i >= 0; ) {
      v = this.pgon.holes[i];   /* traverse the hole vertices */
      x = (int)(w +this.scale *(v.x -this.xoff) +0.5);
      y = (int)(h -this.scale *(v.y -this.yoff) +0.5);
      g.setColor(Color.black);  /* black outline */
      g.fillRect(x-1, y-2, 3, 5);
      g.fillRect(x-2, y-1, 5, 3);
      g.setColor(Color.red);    /* red interior */
      g.fillRect(x-1, y-1, 3, 3);
    }                           /* draw a circle */
  }  /* paint() */

  /*------------------------------------------------------------------*/
  /** Find the closest vertex to a given point.
   *  @param  x   the x-coordinate of the point
   *  @param  y   the y-coordinate of the point
   *  @return the closest vertex to the given point
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Vertex findVertex (int x, int y)
  {                             /* --- find closest vertex */
    if (this.pgon == null)      /* if there is no pointgon, */
      return null;              /* abort the function */
    return this.pgon.findVertex((x-8)           /this.scale +this.xoff,
           (this.getPreferredSize().height-8-y) /this.scale +this.yoff);
  }  /* findVertex() */

  /*------------------------------------------------------------------*/
  /** Find the closest edge to a given point.
   *  @param  x   the x-coordinate of the point
   *  @param  y   the y-coordinate of the point
   *  @return the closest edge to the given point
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Edge findEdge (int x, int y)
  {                             /* --- find closest edge */
    if (this.pgon == null)      /* if there is no pointgon, */
      return null;              /* abort the function */
    return this.pgon.findEdge((x-8) /this.scale +this.xoff,
           (this.getPreferredSize().height-8-y) /this.scale +this.yoff);
  }  /* findEdge() */

  /*------------------------------------------------------------------*/
  /** Create an image of the panel contents.
   *  @return the created image
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public BufferedImage makeImage ()
  {                             /* --- create an image of contents */
    BufferedImage img;          /* created image */
    Dimension     d;            /* size of panel */

    d   = this.getPreferredSize();
    img = new BufferedImage(d.width, d.height,
            BufferedImage.TYPE_3BYTE_BGR);
    this.paint(img.getGraphics());
    return img;                 /* draw window contents to image */
  }  /* BufferedImage() */      /* and return the image */

}  /* class PGPanel */
